@@ -8,6 +8,7 @@
 //
 // SPDX-License-Identifier: MIT/Apache-2.0
 
+use atomic_refcell::AtomicRefCell;
 use gst::glib;
 use gst::gst_debug;
 use gst::subclass::prelude::*;
@@ -19,15 +20,29 @@ use rav1e::config;
 use rav1e::data;
 use std::sync::Mutex;
 
-const DEFAULT_SPEED_PRESET: u32 = 5;
+#[derive(Debug, PartialEq, Eq, Clone, Copy, glib::Enum)]
+#[repr(u32)]
+#[enum_type(name = "GstRav1eEncTune")]
+pub enum Tune {
+    Psnr,
+    Psychovisual,
+}
+
+const DEFAULT_SPEED_PRESET: u32 = 6;
 const DEFAULT_LOW_LATENCY: bool = false;
 const DEFAULT_MIN_KEY_FRAME_INTERVAL: u64 = 12;
 const DEFAULT_MAX_KEY_FRAME_INTERVAL: u64 = 240;
+const DEFAULT_SWITCH_FRAME_INTERVAL: u64 = 0;
 const DEFAULT_BITRATE: i32 = 0;
 const DEFAULT_QUANTIZER: usize = 100;
+const DEFAULT_MIN_QUANTIZER: u8 = 0;
 const DEFAULT_TILE_COLS: usize = 0;
 const DEFAULT_TILE_ROWS: usize = 0;
 const DEFAULT_TILES: usize = 0;
+const DEFAULT_RDO_LOOKAHEAD_FRAMES: i32 = -1;
+const DEFAULT_TUNE: Tune = Tune::Psychovisual;
+const DEFAULT_RESERVOIR_FRAME_DELAY: i32 = i32::MIN;
+const DEFAULT_ERROR_RESILIENT: bool = false;
 const DEFAULT_THREADS: usize = 0;
 
 #[derive(Debug, Clone, Copy)]
@@ -36,11 +51,17 @@ struct Settings {
     low_latency: bool,
     min_key_frame_interval: u64,
     max_key_frame_interval: u64,
+    switch_frame_interval: u64,
     bitrate: i32,
     quantizer: usize,
+    min_quantizer: u8,
     tile_cols: usize,
     tile_rows: usize,
     tiles: usize,
+    rdo_lookahead_frames: i32,
+    tune: Tune,
+    reservoir_frame_delay: i32,
+    error_resilient: bool,
     threads: usize,
 }
 
@@ -51,11 +72,17 @@ impl Default for Settings {
             low_latency: DEFAULT_LOW_LATENCY,
             min_key_frame_interval: DEFAULT_MIN_KEY_FRAME_INTERVAL,
             max_key_frame_interval: DEFAULT_MAX_KEY_FRAME_INTERVAL,
+            switch_frame_interval: DEFAULT_SWITCH_FRAME_INTERVAL,
             bitrate: DEFAULT_BITRATE,
             quantizer: DEFAULT_QUANTIZER,
+            min_quantizer: DEFAULT_MIN_QUANTIZER,
             tile_cols: DEFAULT_TILE_COLS,
             tile_rows: DEFAULT_TILE_ROWS,
             tiles: DEFAULT_TILES,
+            rdo_lookahead_frames: DEFAULT_RDO_LOOKAHEAD_FRAMES,
+            tune: DEFAULT_TUNE,
+            reservoir_frame_delay: DEFAULT_RESERVOIR_FRAME_DELAY,
+            error_resilient: DEFAULT_ERROR_RESILIENT,
             threads: DEFAULT_THREADS,
         }
     }
@@ -67,6 +94,13 @@ enum Context {
 }
 
 impl Context {
+    fn container_sequence_header(&self) -> Vec<u8> {
+        match self {
+            Context::Eight(ref context) => context.container_sequence_header(),
+            Context::Sixteen(ref context) => context.container_sequence_header(),
+        }
+    }
+
     fn receive_packet(
         &mut self,
     ) -> Result<(data::FrameType, u64, u32, Vec<u8>), data::EncoderStatus> {
@@ -98,83 +132,76 @@ impl Context {
 
     fn send_frame(
         &mut self,
-        in_frame: Option<(u32, &gst_video::VideoFrameRef<&gst::BufferRef>)>,
+        frame_number: u32,
+        in_frame: &gst_video::VideoFrameRef<&gst::BufferRef>,
         force_keyframe: bool,
     ) -> Result<(), data::EncoderStatus> {
         match self {
             Context::Eight(ref mut context) => {
-                if let Some((frame_number, in_frame)) = in_frame {
-                    let mut enc_frame = context.new_frame();
-                    enc_frame.planes[0].copy_from_raw_u8(
-                        in_frame.plane_data(0).unwrap(),
-                        in_frame.plane_stride()[0] as usize,
+                let mut enc_frame = context.new_frame();
+                enc_frame.planes[0].copy_from_raw_u8(
+                    in_frame.plane_data(0).unwrap(),
+                    in_frame.plane_stride()[0] as usize,
+                    1,
+                );
+
+                if in_frame.n_planes() > 1 {
+                    enc_frame.planes[1].copy_from_raw_u8(
+                        in_frame.plane_data(1).unwrap(),
+                        in_frame.plane_stride()[1] as usize,
+                        1,
+                    );
+                    enc_frame.planes[2].copy_from_raw_u8(
+                        in_frame.plane_data(2).unwrap(),
+                        in_frame.plane_stride()[2] as usize,
                         1,
                     );
-
-                    if in_frame.n_planes() > 1 {
-                        enc_frame.planes[1].copy_from_raw_u8(
-                            in_frame.plane_data(1).unwrap(),
-                            in_frame.plane_stride()[1] as usize,
-                            1,
-                        );
-                        enc_frame.planes[2].copy_from_raw_u8(
-                            in_frame.plane_data(2).unwrap(),
-                            in_frame.plane_stride()[2] as usize,
-                            1,
-                        );
-                    }
-
-                    context.send_frame((
-                        enc_frame,
-                        Some(rav1e::data::FrameParameters {
-                            frame_type_override: if force_keyframe {
-                                rav1e::prelude::FrameTypeOverride::Key
-                            } else {
-                                rav1e::prelude::FrameTypeOverride::No
-                            },
-                            opaque: Some(rav1e::prelude::Opaque::new(frame_number)),
-                        }),
-                    ))
-                } else {
-                    context.send_frame(None)
                 }
+
+                context.send_frame((
+                    enc_frame,
+                    Some(rav1e::data::FrameParameters {
+                        frame_type_override: if force_keyframe {
+                            rav1e::prelude::FrameTypeOverride::Key
+                        } else {
+                            rav1e::prelude::FrameTypeOverride::No
+                        },
+                        opaque: Some(rav1e::prelude::Opaque::new(frame_number)),
+                    }),
+                ))
             }
             Context::Sixteen(ref mut context) => {
-                if let Some((frame_number, in_frame)) = in_frame {
-                    let mut enc_frame = context.new_frame();
-                    enc_frame.planes[0].copy_from_raw_u8(
-                        in_frame.plane_data(0).unwrap(),
-                        in_frame.plane_stride()[0] as usize,
+                let mut enc_frame = context.new_frame();
+                enc_frame.planes[0].copy_from_raw_u8(
+                    in_frame.plane_data(0).unwrap(),
+                    in_frame.plane_stride()[0] as usize,
+                    2,
+                );
+
+                if in_frame.n_planes() > 1 {
+                    enc_frame.planes[1].copy_from_raw_u8(
+                        in_frame.plane_data(1).unwrap(),
+                        in_frame.plane_stride()[1] as usize,
+                        2,
+                    );
+                    enc_frame.planes[2].copy_from_raw_u8(
+                        in_frame.plane_data(2).unwrap(),
+                        in_frame.plane_stride()[2] as usize,
                         2,
                     );
-
-                    if in_frame.n_planes() > 1 {
-                        enc_frame.planes[1].copy_from_raw_u8(
-                            in_frame.plane_data(1).unwrap(),
-                            in_frame.plane_stride()[1] as usize,
-                            2,
-                        );
-                        enc_frame.planes[2].copy_from_raw_u8(
-                            in_frame.plane_data(2).unwrap(),
-                            in_frame.plane_stride()[2] as usize,
-                            2,
-                        );
-                    }
-
-                    context.send_frame((
-                        enc_frame,
-                        Some(rav1e::data::FrameParameters {
-                            frame_type_override: if force_keyframe {
-                                rav1e::prelude::FrameTypeOverride::Key
-                            } else {
-                                rav1e::prelude::FrameTypeOverride::No
-                            },
-                            opaque: Some(rav1e::prelude::Opaque::new(frame_number)),
-                        }),
-                    ))
-                } else {
-                    context.send_frame(None)
                 }
+
+                context.send_frame((
+                    enc_frame,
+                    Some(rav1e::data::FrameParameters {
+                        frame_type_override: if force_keyframe {
+                            rav1e::prelude::FrameTypeOverride::Key
+                        } else {
+                            rav1e::prelude::FrameTypeOverride::No
+                        },
+                        opaque: Some(rav1e::prelude::Opaque::new(frame_number)),
+                    }),
+                ))
             }
         }
     }
@@ -194,7 +221,7 @@ struct State {
 
 #[derive(Default)]
 pub struct Rav1Enc {
-    state: Mutex<Option<State>>,
+    state: AtomicRefCell<Option<State>>,
     settings: Mutex<Settings>,
 }
 
@@ -251,6 +278,15 @@ impl ObjectImpl for Rav1Enc {
                     DEFAULT_MAX_KEY_FRAME_INTERVAL,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecUInt64::new(
+                    "switch-frame-interval",
+                    "Switch Frame Interval",
+                    "Switch Frame Interval",
+                    0,
+                    std::u64::MAX,
+                    DEFAULT_SWITCH_FRAME_INTERVAL,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
                 glib::ParamSpecInt::new(
                     "bitrate",
                     "Bitrate",
@@ -269,6 +305,15 @@ impl ObjectImpl for Rav1Enc {
                     DEFAULT_QUANTIZER as u32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecUInt::new(
+                    "min-quantizer",
+                    "Min Quantizer",
+                    "Min Quantizer",
+                    0,
+                    std::u8::MAX as u32,
+                    DEFAULT_MIN_QUANTIZER as u32,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
                 glib::ParamSpecUInt::new(
                     "tile-cols",
                     "Tile Cols",
@@ -296,6 +341,39 @@ impl ObjectImpl for Rav1Enc {
                     DEFAULT_TILES as u32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecInt::new(
+                    "rdo-lookahead-frames",
+                    "RDO Lookahead Frames",
+                    "RDO Lookahead Frames",
+                    -1,
+                    std::i32::MAX,
+                    DEFAULT_RDO_LOOKAHEAD_FRAMES,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecEnum::new(
+                    "tune",
+                    "Tune",
+                    "Tune",
+                    Tune::static_type(),
+                    DEFAULT_TUNE as i32,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecInt::new(
+                    "reservoir-frame-delay",
+                    "Reservoir Frame Delay",
+                    "Reservoir Frame Delay",
+                    std::i32::MIN,
+                    std::i32::MAX,
+                    DEFAULT_RESERVOIR_FRAME_DELAY,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecBoolean::new(
+                    "error-resilient",
+                    "Error Resilient",
+                    "Error Resilient",
+                    DEFAULT_ERROR_RESILIENT,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
                 glib::ParamSpecUInt::new(
                     "threads",
                     "Threads",
@@ -335,6 +413,10 @@ impl ObjectImpl for Rav1Enc {
                 let mut settings = self.settings.lock().unwrap();
                 settings.max_key_frame_interval = value.get().expect("type checked upstream");
             }
+            "switch-frame-interval" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.switch_frame_interval = value.get().expect("type checked upstream");
+            }
             "bitrate" => {
                 let mut settings = self.settings.lock().unwrap();
                 settings.bitrate = value.get().expect("type checked upstream");
@@ -343,6 +425,10 @@ impl ObjectImpl for Rav1Enc {
                 let mut settings = self.settings.lock().unwrap();
                 settings.quantizer = value.get::<u32>().expect("type checked upstream") as usize;
             }
+            "min-quantizer" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.min_quantizer = value.get::<u32>().expect("type checked upstream") as u8;
+            }
             "tile-cols" => {
                 let mut settings = self.settings.lock().unwrap();
                 settings.tile_cols = value.get::<u32>().expect("type checked upstream") as usize;
@@ -355,6 +441,22 @@ impl ObjectImpl for Rav1Enc {
                 let mut settings = self.settings.lock().unwrap();
                 settings.tiles = value.get::<u32>().expect("type checked upstream") as usize;
             }
+            "rdo-lookahead-frames" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.rdo_lookahead_frames = value.get::<i32>().expect("type checked upstream");
+            }
+            "tune" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.tune = value.get::<Tune>().expect("type checked upstream");
+            }
+            "reservoir-frame-delay" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.reservoir_frame_delay = value.get::<i32>().expect("type checked upstream");
+            }
+            "error-resilient" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.error_resilient = value.get::<bool>().expect("type checked upstream");
+            }
             "threads" => {
                 let mut settings = self.settings.lock().unwrap();
                 settings.threads = value.get::<u32>().expect("type checked upstream") as usize;
@@ -381,6 +483,10 @@ impl ObjectImpl for Rav1Enc {
                 let settings = self.settings.lock().unwrap();
                 settings.max_key_frame_interval.to_value()
             }
+            "switch-frame-interval" => {
+                let settings = self.settings.lock().unwrap();
+                settings.switch_frame_interval.to_value()
+            }
             "bitrate" => {
                 let settings = self.settings.lock().unwrap();
                 settings.bitrate.to_value()
@@ -389,6 +495,10 @@ impl ObjectImpl for Rav1Enc {
                 let settings = self.settings.lock().unwrap();
                 (settings.quantizer as u32).to_value()
             }
+            "min-quantizer" => {
+                let settings = self.settings.lock().unwrap();
+                (settings.min_quantizer as u32).to_value()
+            }
             "tile-cols" => {
                 let settings = self.settings.lock().unwrap();
                 (settings.tile_cols as u32).to_value()
@@ -401,6 +511,22 @@ impl ObjectImpl for Rav1Enc {
                 let settings = self.settings.lock().unwrap();
                 (settings.tiles as u32).to_value()
             }
+            "rdo-lookahead-frames" => {
+                let settings = self.settings.lock().unwrap();
+                (settings.rdo_lookahead_frames as i32).to_value()
+            }
+            "tune" => {
+                let settings = self.settings.lock().unwrap();
+                settings.tune.to_value()
+            }
+            "reservoir-frame-delay" => {
+                let settings = self.settings.lock().unwrap();
+                settings.reservoir_frame_delay.to_value()
+            }
+            "error-resilient" => {
+                let settings = self.settings.lock().unwrap();
+                settings.error_resilient.to_value()
+            }
             "threads" => {
                 let settings = self.settings.lock().unwrap();
                 (settings.threads as u32).to_value()
@@ -483,7 +609,7 @@ impl ElementImpl for Rav1Enc {
 
 impl VideoEncoderImpl for Rav1Enc {
     fn stop(&self, _element: &Self::Type) -> Result<(), gst::ErrorMessage> {
-        *self.state.lock().unwrap() = None;
+        *self.state.borrow_mut() = None;
 
         Ok(())
     }
@@ -512,7 +638,6 @@ impl VideoEncoderImpl for Rav1Enc {
 
         let settings = self.settings.lock().unwrap();
 
-        // TODO: More properties, HDR information
         let cfg = config::Config::new()
             .with_encoder_config(config::EncoderConfig {
                 width: video_info.width() as usize,
@@ -574,6 +699,22 @@ impl VideoEncoderImpl for Rav1Enc {
                         gst_video::VideoTransferFunction::Bt202012 => {
                             color::TransferCharacteristics::BT2020_12Bit
                         }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoTransferFunction::Bt202010 => {
+                            color::TransferCharacteristics::BT2020_10Bit
+                        }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoTransferFunction::Smpte2084 => {
+                            color::TransferCharacteristics::SMPTE2084
+                        }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoTransferFunction::AribStdB67 => {
+                            color::TransferCharacteristics::HLG
+                        }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoTransferFunction::Bt601 => {
+                            color::TransferCharacteristics::BT601
+                        }
                         gst_video::VideoTransferFunction::Gamma18
                         | gst_video::VideoTransferFunction::Gamma20
                         | gst_video::VideoTransferFunction::Gamma22
@@ -591,6 +732,18 @@ impl VideoEncoderImpl for Rav1Enc {
                         }
                         gst_video::VideoColorPrimaries::Film => color::ColorPrimaries::GenericFilm,
                         gst_video::VideoColorPrimaries::Bt2020 => color::ColorPrimaries::BT2020,
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoColorPrimaries::Smptest428 => color::ColorPrimaries::XYZ,
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoColorPrimaries::Smpterp431 => {
+                            color::ColorPrimaries::SMPTE431
+                        }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoColorPrimaries::Smpteeg432 => {
+                            color::ColorPrimaries::SMPTE432
+                        }
+                        #[cfg(feature = "hdr")]
+                        gst_video::VideoColorPrimaries::Ebu3213 => color::ColorPrimaries::EBU3213,
                         gst_video::VideoColorPrimaries::Adobergb | _ => {
                             color::ColorPrimaries::Unspecified
                         }
@@ -602,6 +755,60 @@ impl VideoEncoderImpl for Rav1Enc {
                         matrix_coefficients: matrix,
                     })
                 },
+                mastering_display: {
+                    #[cfg(feature = "hdr")]
+                    {
+                        state
+                            .caps()
+                            .and_then(|caps| {
+                                gst_video::VideoMasteringDisplayInfo::from_caps(caps).ok()
+                            })
+                            .map(|info| rav1e::prelude::color::MasteringDisplay {
+                                primaries: [
+                                    rav1e::prelude::color::ChromaticityPoint {
+                                        x: info.display_primaries()[0].x,
+                                        y: info.display_primaries()[0].y,
+                                    },
+                                    rav1e::prelude::color::ChromaticityPoint {
+                                        x: info.display_primaries()[1].x,
+                                        y: info.display_primaries()[1].y,
+                                    },
+                                    rav1e::prelude::color::ChromaticityPoint {
+                                        x: info.display_primaries()[2].x,
+                                        y: info.display_primaries()[2].y,
+                                    },
+                                ],
+                                white_point: rav1e::prelude::color::ChromaticityPoint {
+                                    x: info.white_point().x,
+                                    y: info.white_point().y,
+                                },
+                                max_luminance: info.max_display_mastering_luminance(),
+                                min_luminance: info.min_display_mastering_luminance(),
+                            })
+                    }
+                    #[cfg(not(feature = "hdr"))]
+                    {
+                        None
+                    }
+                },
+                content_light: {
+                    #[cfg(feature = "hdr")]
+                    {
+                        state
+                            .caps()
+                            .and_then(|caps| {
+                                gst_video::VideoContentLightLevel::from_caps(caps).ok()
+                            })
+                            .map(|info| rav1e::prelude::color::ContentLight {
+                                max_content_light_level: info.max_content_light_level(),
+                                max_frame_average_light_level: info.max_frame_average_light_level(),
+                            })
+                    }
+                    #[cfg(not(feature = "hdr"))]
+                    {
+                        None
+                    }
+                },
                 speed_settings: config::SpeedSettings::from_preset(settings.speed_preset as usize),
                 time_base: if video_info.fps() != gst::Fraction::new(0, 1) {
                     data::Rational {
@@ -614,18 +821,35 @@ impl VideoEncoderImpl for Rav1Enc {
                 low_latency: settings.low_latency,
                 min_key_frame_interval: settings.min_key_frame_interval,
                 max_key_frame_interval: settings.max_key_frame_interval,
+                switch_frame_interval: settings.switch_frame_interval,
                 bitrate: settings.bitrate,
                 quantizer: settings.quantizer,
+                min_quantizer: settings.min_quantizer,
                 tile_cols: settings.tile_cols,
                 tile_rows: settings.tile_rows,
                 tiles: settings.tiles,
+                rdo_lookahead_frames: if settings.rdo_lookahead_frames < 0 {
+                    config::SpeedSettings::rdo_lookahead_frames(settings.speed_preset as usize)
+                } else {
+                    settings.rdo_lookahead_frames as usize
+                },
+                tune: match settings.tune {
+                    Tune::Psnr => rav1e::prelude::Tune::Psnr,
+                    Tune::Psychovisual => rav1e::prelude::Tune::Psychovisual,
+                },
+                reservoir_frame_delay: if settings.reservoir_frame_delay == i32::MIN {
+                    None
+                } else {
+                    Some(settings.reservoir_frame_delay)
+                },
+                error_resilient: settings.error_resilient,
                 ..Default::default()
             })
             .with_threads(settings.threads);
         // TODO: RateControlConfig
 
-        *self.state.lock().unwrap() = Some(State {
-            context: if video_info.format_info().depth()[0] > 8 {
+        let context =
+            if video_info.format_info().depth()[0] > 8 {
                 Context::Sixteen(cfg.new_context().map_err(|err| {
                     gst::loggable_error!(CAT, "Failed to create context: {:?}", err)
                 })?)
@@ -633,7 +857,12 @@ impl VideoEncoderImpl for Rav1Enc {
                 Context::Eight(cfg.new_context().map_err(|err| {
                     gst::loggable_error!(CAT, "Failed to create context: {:?}", err)
                 })?)
-            },
+            };
+        let container_sequence_header =
+            gst::Buffer::from_mut_slice(context.container_sequence_header());
+
+        *self.state.borrow_mut() = Some(State {
+            context,
             video_info,
         });
 
@@ -642,6 +871,7 @@ impl VideoEncoderImpl for Rav1Enc {
                 gst::Caps::builder("video/x-av1")
                     .field("stream-format", "obu-stream")
                     .field("alignment", "tu")
+                    .field("codec_data", container_sequence_header)
                     .build(),
                 Some(state),
             )
@@ -656,7 +886,7 @@ impl VideoEncoderImpl for Rav1Enc {
     fn flush(&self, element: &Self::Type) -> bool {
         gst_debug!(CAT, obj: element, "Flushing");
 
-        let mut state_guard = self.state.lock().unwrap();
+        let mut state_guard = self.state.borrow_mut();
         if let Some(ref mut state) = *state_guard {
             state.context.flush();
             while let Ok(_) | Err(data::EncoderStatus::Encoded) = state.context.receive_packet() {
@@ -670,11 +900,8 @@ impl VideoEncoderImpl for Rav1Enc {
     fn finish(&self, element: &Self::Type) -> Result<gst::FlowSuccess, gst::FlowError> {
         gst_debug!(CAT, obj: element, "Finishing");
 
-        let mut state_guard = self.state.lock().unwrap();
+        let mut state_guard = self.state.borrow_mut();
         if let Some(ref mut state) = *state_guard {
-            if let Err(data::EncoderStatus::Failure) = state.context.send_frame(None, false) {
-                return Err(gst::FlowError::Error);
-            }
             state.context.flush();
             self.output_frames(element, state)?;
         }
@@ -687,7 +914,7 @@ impl VideoEncoderImpl for Rav1Enc {
         element: &Self::Type,
         frame: gst_video::VideoCodecFrame,
     ) -> Result<gst::FlowSuccess, gst::FlowError> {
-        let mut state_guard = self.state.lock().unwrap();
+        let mut state_guard = self.state.borrow_mut();
         let state = state_guard.as_mut().ok_or(gst::FlowError::NotNegotiated)?;
 
         self.output_frames(element, state)?;
@@ -713,7 +940,8 @@ impl VideoEncoderImpl for Rav1Enc {
             })?;
 
         match state.context.send_frame(
-            Some((frame.system_frame_number(), &in_frame)),
+            frame.system_frame_number(),
+            &in_frame,
             frame
                 .flags()
                 .contains(gst_video::VideoCodecFrameFlags::FORCE_KEYFRAME),
@@ -767,6 +995,10 @@ impl Rav1Enc {
                 Err(data::EncoderStatus::Encoded) => {
                     gst_debug!(CAT, obj: element, "Encoded but not output frame yet",);
                 }
+                Err(data::EncoderStatus::NeedMoreData) => {
+                    gst_debug!(CAT, obj: element, "Encoded but need more data",);
+                    return Ok(gst::FlowSuccess::Ok);
+                }
                 Err(data::EncoderStatus::Failure) => {
                     gst::element_error!(
                         element,
