@@ -5,6 +5,8 @@
 // <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
+//
+// SPDX-License-Identifier: MIT/Apache-2.0
 
 use gst::prelude::*;
 
@@ -121,7 +123,7 @@ fn test_encode(video_info: &gst_video::VideoInfo) {
     let mut h = gst_check::Harness::new("rav1enc");
     {
         let rav1enc = h.element().unwrap();
-        rav1enc.set_property("speed-preset", &10u32).unwrap();
+        rav1enc.set_property("speed-preset", 10u32);
     }
     h.play();
     h.set_src_caps(video_info.to_caps().unwrap());
